pro sxt_sff_composite, index, data, comp_index, comp_data, $
	satpix=satpix,plot=plot,                                  $
	smooth_factor=smooth_factor,sfd=sfd,		        	$
	qtest=qtest
;+
; NAME:
;   sxt_sff_composite
; PURPOSE:
;   Prepare a composite SXT image from 2 or 3 Half or Quarter Res images.
; CALLING SEQUENCE EXAMPLES:
; INPUTS:
;   index = SXT index structure (2 or 3 elements)
;   data  = 3-d data cube (at least 2 images)
;  
;						-- and background subtracted
; OUTPUT:
;   comp_index - the composite index with history
;   comp_data  - the composite image (2D)
;
; OPTIONAL INPUT KEYWORDS:
;   plot   = If set, plot the results as we go.  The results are displayed
;	     on a logarithmic scale.
;   smooth_factor = Boxcar smooth the bleed pixels.  If not defined, the
;	     default is to smooth over 5 pixels.
; MODIFICATION HISTORY:
;      8-Nov-91, L. W. Acton (Wrote composite)
;       (many mods)
;     10-Jan-97, LWA - Corrected calling sequence examples.
;     12-Sep-1998 - S.L.Freeland from Loren Acton's sxt_composite
;                  Stripped for final movie, use sxt_prep internnally 
;                  Starting from sxt_composite 10-Jan-97, LWA 
;-

progverno=1.00*1000
nparams = n_params()
if nparams lt 2 then begin
   doc_library,'sxt_composite'
   return
endif

;-------------------------
;  Check input parameters
;-------------------------

nimg=data_chk(data,/nimag)
if nimg lt 2 then begin
   print,'***  Error in sxt_composite:  data must be 3-d'
   help,data
   return
endif

if n_elements(index) lt 2 then begin	; Must be at least 2 images
   print,'***  Error in sxt_composite:  index must have minimum length of 2'
   help,index
   return
endif else nimg = nimg < n_elements(index)

if n_elements(sx) eq 0 then sx = indgen(nimg < 3)
nimg = n_elements(sx)
sx = sx(sort(gt_expdur(index(sx))))	; short, medium, long

; Make sure all the images have the same resolution:
if gt_res(index(sx(0))) ne total(gt_res(index(sx)))/ nimg then begin
    message,'The images must all have the same resolution'
    print,'Resolutions present = ',gt_res(index(sx),/str)
    return
endif

if not data_chk(data,/type) eq 1 then begin 
 box_message,['This routine assum "raw" sfr/spr input', $
         'Use <sxt_composite.pro> for user customization']
endif 

sxt_prep,index(sx),data(*,*,sx),comp_index, comp_data,unc_data,satpix, $
   /register,/float,/dc_orbit_correct, /vignette , $
   ref_image=index(sx(0))

s_sat=indgen(nimg)                ; data arrays now ordered
sx=indgen(nimg)
t0=gt_expdur(comp_index)
; ---------------------------------------------------------------------------
; Composite Algorithm
; ---------------------------------------------------------------------------
; Algorithm for 3 image case:
;
;1.  Replace the saturated pixels in the medium exposure with the
;    corresponding pixels in the short exposure.
;2.  Replace saturated pixels in long exposure with corresponding pixels
;    from the combined medium/short exposure == cimg (combined image).
;3.  Make a temporary copy of the combined image and smooth with a
;    boxcar width of 5  == timg (temporary image).
;4.  cimg(s2) = timg(s2) where s2 = saturated pixels in the long exposure.
;5.  Let cimg(m1) = short_exposure(m1) where m1 is given by:
;    m1 = where((short_exposure ge 90) and (short_expousre is not saturated) and
;	         (long_exposure is saturated))
;
; Algorithm for 2 image case:
;   Same as above except Step 1 is omitted.
; ---------------------------------------------------------------------------

; Set up the output array (cimg2) and working array (cimg1):
xsize=data_chk(comp_data,/nx)
ysize=data_chk(comp_data,/ny)
typx =data_chk(comp_data,/type) > 3	; Long or float if float on input
cimg1 = make_array(xsize,ysize,type=typx)	; short or combined short/med
cimg2 = make_array(xsize,ysize,type=typx)	; Final image

; ---------------------------------------------------------------------------
; This sets up m1 (see step 5).
; m1 = where((short_exposure ge 90) and (short_exposure is not saturated) and
;            (long_exposure is saturated))
; ---------------------------------------------------------------------------
cimg1(0,0) = comp_data(*,*,nimg-2) 	; Short (3- or 2-image)
cimg2(0,0) = comp_data(*,*,nimg-1)	; Long (3-image) or Med (2-image)
m1 = where((comp_data(*,*,0) ge 90) and 			$; High signal in short, 
		(satpix(*,*,s_sat(0)) eq 0) and 	$; but short not saturated, &
		(satpix(*,*,s_sat(nimg-1)) eq 1),ngood)	 ; saturated in long exposure
   if ngood gt 0 then  m1_data = (comp_data(*,*,0))(m1)	 ; Use the short exposure
;Removed >0 in following 2 statements.  LWA 5/08/95.
   cimg1(0,0) = comp_data(*,*,sx(nimg-2))
   cimg2(0,0) = comp_data(*,*,sx(nimg-1))
stop
   m1 = where((comp_data(*,*,sx(0)) ge 90) and		$; High signal in short,
		(satpix(*,*,s_sat(0)) eq 0) and 	$; but short not saturated, &
		(satpix(*,*,s_sat(nimg-1)) eq 1),ngood)	 ; saturated in long exposure
   if ngood gt 0 then  m1_data = (comp_data(*,*,sx(0)))(m1)	 ; Use the short exposure

; ---------------------------------------------------------------------------
; This is Step 1 for 3-image case:  Insert shortest into the medium exposure
; ---------------------------------------------------------------------------
if nimg eq 3 then begin
  s1 = where(satpix(*,*,s_sat(nimg-2)),npix)	; s1 = saturated pixels in med. exposure
  if npix gt 0 then begin
    if n_elements(img) ne 0 then 	$
	cimg1(s1) = ((comp_data(*,*,nimg-3))(s1)    * t0(nimg-2)/t0(nimg-3)+.5) else $
;Removed >0 in following statement.  LWA 5/08/95.
	cimg1(s1) = (((comp_data(*,*,sx(nimg-3)))(s1))*t0(nimg-2)/t0(nimg-3)+.5)
  endif
endif
delvarx,img					; Finished with img array

; ---------------------------------------------------------------------------
; This is Step 2: Replace saturated pixels in long exposure with 
;                 short (2-image) or combined med/short (3-image)
; ---------------------------------------------------------------------------
s2 = where(satpix(*,*,s_sat(nimg-1)),npix)	; Saturated pixels in long exposure
if npix gt 0 then 		$		; Replace saturated pixels in long exposure
  cimg2(s2) = cimg1(s2) * t0(nimg-1) / t0(nimg-2) + .5
delvarx,cimg1					; Finished with short/med array

; ---------------------------------------------------------------------------
; This section represents Steps 3 and 4:
; ---------------------------------------------------------------------------
; Make smoothed array and use it to replace bleed pixel
if n_elements(smooth_factor) eq 0 then smooth_factor = 5
if smooth_factor gt 1 then begin
  if fix(smooth_factor/2)*2 eq smooth_factor then message,'Warning: Smooth_factor is not an odd number = '+strtrim(smooth_factor,2),/info
  if (npix gt 0) then begin			;MDM put conditional 1-Dec-93
    timg = smooth(cimg2,smooth_factor)		; Boxcar smooth over smooth_factor pixels
    cimg2(s2) = timg(s2)
  endif						; npix gt 0
endif

; ---------------------------------------------------------------------------
; Implement Step 5:
; ---------------------------------------------------------------------------
if ngood gt 0 then cimg2(m1) = (m1_data * t0(nimg-1)/t0(0) + .5)
if keyword_set(qtest) then stop 

;---------------------------------------------------------------------------
; This section builds the comp_index structure.
; ** Still need to include index.his.Q_COMPRESSION once I find out how
;	is it used.   LWA 2-Nov-96 **
;---------------------------------------------------------------------------
comp_index=index(sx(nimg-1))			;Longest exposure.
his_index, /enable
if nimg eq 2 then begin
   time_compos=[gt_time(index(sx(0))),gt_time(index(sx(1))),0]
   day_compos=[gt_day(index(sx(0))),gt_day(index(sx(1))),0]
endif else begin
   time_compos=[gt_time(index(sx(0))),gt_time(index(sx(1))),$
	gt_time(index(sx(2)))]
   day_compos=[gt_day(index(sx(0))),gt_day(index(sx(1))),gt_day(index(sx(2)))]
endelse

his_index, comp_index, 0, 'q_composite', progverno*1000
his_index, comp_index, 0, 'time_compos', time_compos
his_index, comp_index, 0, 'day_compos', day_compos

comp_data=temporary(cimg2)
return 
end

