pro sxt_sff_composite,index,data,comp_index,comp_data, $
	smooth_factor=smooth_factor,                   $
        sxt_kleen=sxt_kleen, vignette=vignette
;+
; NAME:
;   sxt_sff_composite
;
; PURPOSE:
;   Prepare a composite SXT image from 2 or 3 Half or Quarter Res images.
;
; CALLING SEQUENCE EXAMPLES:
;
; INPUTS:
;   index = SXT index structure
;   data  = 3-d data cube (at least 2 images)
;
;
; OUTPUT:
;   comp_index - the adjusted index of the composite (history added)
;   comp_data  - the composite data array
;
;
; OPTIONAL INPUT KEYWORDS:
;
;   smooth_factor = Boxcar smooth the bleed pixels.  If not defined, the
;	     default is to smooth over 5 pixels.
;   _extra - all others passed through to sxt_prep
;
; MODIFICATION HISTORY:
;   15-sep-98  S.L.Freeland - from LWA 'sxt_composite.pro'
;              stripped version using 'new technology'
;              Calls sxt_prep internally
;
;    Previous History:
;    8-Nov-91, L. W. Acton (Wrote composite)
;   19-May-92, LWA  (Several more mods in betwee)
;   13-may-93, J. R. Lemen, Re-wrote.  Calls sxt_satpix
;    1-Dec-93, MDM - Corrected error with the use of "s2" (when it is -1)
;		- Modification on selection of replacement pixels.
;		  It used to check (a) that it was high signal (ge 90)
;				   (b) that it was not saturated
;		  Added check so that it replaced pixels in the longer 
;		  exposure only when the longer exposure was saturated
;    15-Dec-93, SLF - Fixed typo (stapix->satpix)
;    22-feb-94, JRL - Corrected the algorithm; 
;		    - added smooth_factor, sfd, dc_interpolate, register keywords
;    17-Apr-94, LWA - Fixed /sfd scaling option (1000.*  was needed)
;		    - Changed /plot to plot raw input images.
;		    - Added RESTRICTIONS statement to header.
;		    - Added despike=despike keyword.
;    23-jun-94, JRL - Output is floating if input is a fltarr;
;		    - Call sfd_comp with index=index to normalize properly
;    08-May-95, LWA - Fixed to return negative values.
;    02-Nov-96, LWA - Added output of index with history record.
;		    - Made dc_data a keyword intput.
;		    - Made dc_interpolate the default is dc_data not input.
;    04-Nov-96, LWA - Removed forgotten diagnostic.
;    10-Jan-97, LWA - Corrected calling sequence examples.
;-

progverno=4.00*1000
nparams = n_params()
if nparams lt 4 then begin
   box_message,'IDL> sxt_sff_composite,index,data,compindex,compdata
   return
endif

;-------------------------
;  Check input parameters
;-------------------------
;

sz0 = size(data)			; Data must be 3-d
nimg=data_chk(data,/nimage)

if nimg lt 2 or n_elements(index) lt 2 then begin
   print,'***  Error in sxt_composite:  index,data must be 3-d'
   return
endif

exporder=sort(gt_expdur(index))

; ----- the folling call orders output short->long
sxt_prep,index(exporder),data(*,*,exporder),comp_index,comp_data,unc,satpix, $
  /register,/float,/dc_orbit_correct,ref_image=index(exporder(nimg-1)), $
  vignette=vignette, sxt_kleen=sxt_kleen

t0 = gt_expdur(comp_index)

xsize = data_chk(comp_data,/nx)
ysize = data_chk(comp_data,/ny)

; ---------------------------------------------------------------------------
; Composite Algorithm
; ---------------------------------------------------------------------------
; Algorithm for 3 image case:
;
;1.  Replace the saturated pixels in the medium exposure with the
;    corresponding pixels in the short exposure.
;2.  Replace saturated pixels in long exposure with corresponding pixels
;    from the combined medium/short exposure == cimg (combined image).
;3.  Make a temporary copy of the combined image and smooth with a
;    boxcar width of 5  == timg (temporary image).
;4.  cimg(s2) = timg(s2) where s2 = saturated pixels in the long exposure.
;5.  Let cimg(m1) = short_exposure(m1) where m1 is given by:
;    m1 = where((short_exposure ge 90) and (short_expousre is not saturated) and
;	         (long_exposure is saturated))
;
; Algorithm for 2 image case:
;   Same as above except Step 1 is omitted.
; ---------------------------------------------------------------------------

; ---------------------------------------------------------------------------
; This sets up m1 (see step 5).
; m1 = where((short_exposure ge 90) and (short_exposure is not saturated) and
;            (long_exposure is saturated))
; ---------------------------------------------------------------------------

   cimg1 = comp_data(*,*,nimg-2) 	; Short (3- or 2-image)
   cimg2 = comp_data(*,*,nimg-1)		; Long (3-image) or Med (2-image)
   m1 = where((comp_data(*,*,0) ge 90) and 			$; High signal in short, 
		(satpix(*,*,0) eq 0) and 	$; but short not saturated, &
		(satpix(*,*,nimg-1) eq 1),ngood)	 ; saturated in long exposure
   if ngood gt 0 then  m1_data = (comp_data(*,*,0))(m1)	 ; Use the short exposure

; ---------------------------------------------------------------------------
; This is Step 1 for 3-image case:  Insert shortest into the medium exposure
; ---------------------------------------------------------------------------
if nimg eq 3 then begin
  s1 = where(satpix(*,*,nimg-2),npix)	; s1 = saturated pixels in med. exposure
  if npix gt 0 then $    	
	cimg1(s1) = ((comp_data(*,*,nimg-3))(s1)    * t0(nimg-2)/t0(nimg-3)+.5)
  
endif
delvarx,img					; Finished with img array

; ---------------------------------------------------------------------------
; This is Step 2: Replace saturated pixels in long exposure with 
;                 short (2-image) or combined med/short (3-image)
; ---------------------------------------------------------------------------
s2 = where(satpix(*,*,nimg-1),npix)	; Saturated pixels in long exposure
if npix gt 0 then 		$	; Replace saturated pixels in long exposure
  cimg2(s2) = cimg1(s2) * t0(nimg-1) / t0(nimg-2) + .5
delvarx,cimg1					; Finished with short/med array

; ---------------------------------------------------------------------------
; This section represents Steps 3 and 4:
; ---------------------------------------------------------------------------
; Make smoothed array and use it to replace bleed pixel
if n_elements(smooth_factor) eq 0 then smooth_factor = 5
if smooth_factor gt 1 then begin
  if fix(smooth_factor/2)*2 eq smooth_factor then message,'Warning: Smooth_factor is not an odd number = '+strtrim(smooth_factor,2),/info
  if (npix gt 0) then begin			;MDM put conditional 1-Dec-93
    timg = smooth(cimg2,smooth_factor)		; Boxcar smooth over smooth_factor pixels
    cimg2(s2) = timg(s2)
  endif						; npix gt 0
endif

; ---------------------------------------------------------------------------
; Implement Step 5:
; ---------------------------------------------------------------------------
if ngood gt 0 then cimg2(m1) = (m1_data * t0(nimg-1)/t0(0) + .5)

;---------------------------------------------------------------------------
; This section builds the comp_index structure.
; ** Still need to include index.his.Q_COMPRESSION once I find out how
;	is it used.   LWA 2-Nov-96 **
;---------------------------------------------------------------------------
comp_index=comp_index(nimg-1)		;Longest exposure.
his_index, /enable
time_compos=gt_time(comp_index)
day_compos =gt_day(comp_index)

his_index, comp_index, 0, 'q_composite', progverno*1000
his_index, comp_index, 0, 'time_compos', time_compos
his_index, comp_index, 0, 'day_compos', day_compos

comp_data=cimg2
return
end

